/****************************************************************************/
/*                                                                          */ 
/* eze_deduce (ver. 0.01): Hierarchical Classification Learner              */ 
/*                                                                          */ 
/* Author: Yutaka Sasaki, Toyota Technological Institute                    */ 
/*         (Yutaka.Sasaki<AT>toyota-ti.ac.jp)                               */ 
/*                                                                          */
/*   Release Date: March 27, 2016                                           */
/*                                                                          */
/*   Acknowledgment: This software development has been partly supported    */
/*                   by Japan's MEXT/JSPS KAKENHI Grant Number 25330271.    */
/*                                                                          */ 
/* Licensed under the GNU General Public License, Version 2 (the "License") */
/* You may not use this file except in compliance with the License.         */
/* You may obtain a copy of the License at                                  */
/*                                                                          */
/*      http://http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt         */
/*                                                                          */
/* Unless required by applicable law or agreed to in writing, software      */
/* distributed under the License is distributed on an "AS IS" BASIS,        */
/* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. */
/* See the License for the specific language governing permissions and      */
/* limitations under the License.                                           */
/*                                                                          */ 
/****************************************************************************/

/* NB: This is not a good C++ code. This is C implementation with C++ data  */
/* type extensions. Minimum Object-Oriented programming has been involved.  */

/* Header files */
#include <fstream>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <map>
#include <set>
#include <vector>
#include <queue>
#include <unistd.h>
#include "eze-0.01.h"

/* Global variables */
char version[] = "eze_deduce (ver. 0.01): Hierarchical Classification System, Yutaka Sasaki, TTI, Japan.\n";
float* W;                                       /* weight vector */
std::vector< std::vector< std::pair< int, float > > >  X;  /* sample list */
float B;                                        /* SVM bias */
int max_findex = MAX_FINDEX;                    /* default largest feature id */
std::multimap< int, int > child;                /* parent to child */
std::map< int, std::map< int, float > > d2cat;  /* data to class */
char* test_file;                                /* test file name */
char* hierarchy_file;                           /* hierarchy file name */
char model_dir[NAME_LEN]=MODEL;                 /* model directory name */
char result_file[NAME_LEN]=RESULT;              /* result file name */
int verbose = 0;                                /* verbose option flag */
int accs_bias = 1;                              /* ACCS bias option flag */
int logistic  = 0;                              /* logistic pruning option flag */
std::string root = ROOT;
int root_id = 0;
std::multimap< int, int >::iterator cciter;     /* checking iterator */
bool* ccheck;                                   /* already explored or not */
std::map< int, std::string > i2cat;             /* category id to category string */
std::map< std::string, int > cat2i;             /* category string to category id */
int cat_serial = 1;

typedef std::pair<int, int> pair_type;

struct mycomp {
   bool operator()(const pair_type& x, const pair_type& y) const {
     return x.second < y.second;
   }
};

/* Function prototypes */
int read_model_file ( char fname[] );
int read_hierarchy_file ( char fname[] );
int read_test_file ( char fname[] );
bool traverse ( int depth, int node, std::vector< int > *dat_list, std::vector< float > *score_list );
void usage( char* command );
float dot_product( int datid );
void write_class( std::vector< int > *v, char* fname );
void read_option (int argc, char *argv[]);
float range01 ( float x );
float clip ( float x );
float sigmoid ( float x );

/* Main Program */
int main(int argc, char *argv[]) {

  /* Local variables */
  int num_dat;                     /* #data */
  char* command;                   /* command */
  std::vector< int > dat_list;     /* data list */ 
  std::vector< float > score_list; /* scoree list */ 
  int result;                      /* file read result */ 

  /* Read command line loption */
  command = argv[0];
  read_option( argc, argv );
  if ( test_file == '\0' ) { usage(command); exit(1); }  
  if ( hierarchy_file == '\0' ) { usage(command); exit(1); }  

  /* Pring header */
  printf( version );
  printf( "*** Hierarchy file name = %s\n",hierarchy_file );
  printf( "*** Test file name = %s\n",test_file );
  printf( "*** Model directory = %s\n",model_dir );
  printf( "*** Result file name = %s\n",result_file );
  printf( "*** Root node ID = %s\n", root.c_str() );

  /* Initialization */
  i2cat[0] = root;
  cat2i[root] = 0;

  /* Read files */
  fprintf( stderr, "Reading hierarchy file %s ...\n",hierarchy_file );
  result = read_hierarchy_file ( hierarchy_file );
  fprintf( stderr, "Done (%d).\n",result );

  ccheck = new bool[cat_serial];
  for( int i=0; i< cat_serial ; i++ ) ccheck[i] = false;

  fprintf( stderr, "Reading test file %s ...\n",test_file );
  num_dat = read_test_file ( test_file );
  fprintf( stderr, "Done. %d samples read.\n",num_dat );
  
  W = new float[max_findex+2];
  for( int i=0; i< num_dat ; i++ ) 
    { dat_list.push_back(i); score_list.push_back(1.0); }
  fprintf( stderr, "dat# = %d, %d\n", num_dat, (int)(dat_list.size()) );

  /* Traverse hierarchy */
  traverse( 1, root_id, &dat_list, &score_list );

  /* Write results to a result file */
  write_class( &dat_list, result_file );

  /* clean up */
  delete[] W;
}

/*****************************************************************************/
/*  traverse                                                                 */
/*    (Arguments) depth: current depth                                       */
/*                node: target node                                          */
/*                *dat_list: data list                                       */
/*                *score_list: score list                                    */
/*****************************************************************************/ 
bool traverse ( int depth, int node, std::vector< int > *dat_list, std::vector< float > *score_list ) {

  /* local variables */
  int i,c;                                  /* counter */
  float g;                                 /* SVM output */
  float G;                                 /* SVM decision */
  int ccount;                               /* #child */
  int dcount;                               /* #data */
  int result;                               /* result */
  char model_file[200];                     /* model file */
  std::map< int, float > max_g;            /* max SVM output for each data */
  std::map< int, float > max_child;        /* max SVM output child node */
  std::map< int, int > check;               /* max SVM output child node */
  std::multimap< int, std::pair< int, float > > c2dat;  /* class to data */
  std::multimap< int, int >::iterator iter;
  bool rr = false;

  /* initialize */
  fprintf( stderr, "Processing class %s for %d data.\n", i2cat[node].c_str(), (int)(dat_list->size()) );
  for( int i=0 ; i < (int)(dat_list->size()) ; i++ ) { max_child[(*dat_list)[i]]=0; }

  /* leaf node */
  ccount = child.count( node );
  if (verbose) fprintf( stderr, "(%s) %d child nodes found.\n", i2cat[node].c_str(), ccount );
  if ( ccount == 0 ) { 
     dcount = (*dat_list).size();
     for( i=0 ; i<dcount; i++ ) {
       if ( d2cat.count( (*dat_list)[i] ) > 0 ) {
          std::map< int, std::map< int, float > >::iterator diter;
          diter = d2cat.find( (*dat_list)[i] );
          if ( ((*diter).second).count(node) > 0 ) {
            std::map< int, float >::iterator citer;
            citer = ((*diter).second).find( node );
            if ( (*citer).second > (*score_list)[i] ) { continue; }
            else { ((*diter).second).erase( citer );  }
	  }
       }
       //       if ( (*score_list)[i]> 0.20 )
         d2cat[(*dat_list)[i]].insert(std::pair<int,float>(node,(*score_list)[i]));
     } /* end od for */
     return( true ); 
  }

  if ( ccheck[node] ) return(false);
  ccheck[node] = true;

  /* make a decision */
  iter=child.find( node ); 
  for ( c=0 ; c<ccount ; c++ ) {
    sprintf( model_file, "%s/eze-%s-%s.model", model_dir, i2cat[node].c_str(), i2cat[(int)(iter->second)].c_str() );
    if (verbose) fprintf( stderr, "Model file name = %s\n",model_file );
    if (verbose) fprintf( stderr, "Expanding %s-->%s\n", i2cat[node].c_str(), i2cat[(int)(iter->second)].c_str() );
    result = read_model_file ( model_file );
    if ( result == 0 ) { fprintf( stderr, "%s not found!\n",model_file ); iter++; continue; }

    for( int i=0 ; i < (int)(dat_list->size()) ; i++ ) {
      int d = (*dat_list)[i];
      g = dot_product( d );
      if ( accs_bias )  G = g + 2*(*score_list)[i]-1; 
      else G=g; 

      if ( G > 0.0 ) { 
	  c2dat.insert( std::pair<int,std::pair<int,float> > (
	      (*iter).second, std::pair<int,float>(d, (*score_list)[i]*range01(g)) ));
          check[d] = 1; 
        }
        if ( (g>max_g[d])||(max_child[d]==0) ) {
 	   max_g[d] = g; 
           max_child[d] = (*iter).second; 
        }
    }
    iter++; 
  } /* end of for */
  
  /* store scores in c2dat */
  dcount=(*dat_list).size();
  for( i=0 ; i<dcount; i++ ) {
    int d = (*dat_list)[i];
    if ( check[d] == 0 ) {
        c2dat.insert( 
           std::pair<int, std::pair<int, float> >( 
               max_child[d], 
               std::pair<int, float>(d, (*score_list)[i]*range01(max_g[d]) ) ));
    } /* end of if check[d] */
  } /* end of for */

  /* go down to the child level */
  std::priority_queue< pair_type, std::vector<pair_type>, mycomp > que;

  iter=child.find( node ); 
  for ( int i=0 ; i<ccount ; i++ ) {
     pair_type p =  std::pair<int,int>((*iter).second, c2dat.count((*iter).second));
     que.push( p );
     iter++;
  }

  iter=child.find( node ); 
  for ( int i=0 ; i<ccount ; i++ ) {
    std::vector<int> child_dat_list;
    std::vector<float> child_score_list;
    std::multimap< int, std::pair<int, float> >::iterator idat;
    std::multimap< int, int >::iterator thisiter;     /* checking iterator */

    int c = (*iter).second;
    //    int c = (que.top()).first;
    if (verbose) fprintf( stderr, "(%d/%d) class %d with %d data\n", i+1, ccount, c, (int)c2dat.count( c) );
    que.pop();

    dcount=c2dat.count( c ); 
    if ( dcount > 0 ) { 
      idat=c2dat.find( c ); 
      for( int dat=0 ; dat < dcount ; dat++ ) {
          child_dat_list.push_back(   (idat->second).first );
          child_score_list.push_back( (idat->second).second );
          idat++;
      } 
    }
    if ( child_dat_list.size()>0 ) 
      traverse( depth+1, c, &child_dat_list, &child_score_list );
    iter++; 
  } /* end of for */

  ccheck[node] = false;  /* revisit nodes if false */
  return( rr ); 
}

/*****************************************************************************/
/* read_model_file                                                           */
/*    (Arguments) fname: file name                                           */
/*    (Return)    success or not                                             */
/*****************************************************************************/ 
int read_model_file ( char fname[] ) {

  /* local variables */
  std::ifstream ifs;   /* file descriptor */
  int iatt;            /* attribute */
  float fval;          /* value */
  register char* pc;   /* character pointer */
  register char* term; /* term pointer */
  std::string strbuf;  /* string buffer */
  char* buf;           /* char buffer */
  int bufsize;         /* char buffer size */

  /* initialize */
  for( int i=0; i< max_findex ; i++) { W[i] = 0;  }
  if (verbose)  fprintf( stderr, "fname =  %s\n", fname );

  ifs.open( fname );
  if ( ifs.fail() ) { fprintf( stderr, "File not found (%s)\n", fname); return 0; }

  std::getline( ifs,strbuf );
  bufsize = strbuf.size();
  buf = new char[bufsize+1];
  strcpy( buf, strbuf.c_str() );

  /* SVM bais */
  pc = buf;  
  term = pc;

  if ( (pc = strchr(pc,' ')) ) {
    B = atof( term );
    W[0] = B;
    pc++;
    //  printf( "%g\n", B );
   } else {
     B = atof( term );
     W[0] = B;

    delete[] buf;
     ifs.close();
     return 0;
   }

  //  printf( "%g\n", B );

  /* feature list */
  while ( *pc != '\0' ) {
      iatt = 0;
      while ( *pc != ':' ) { iatt = iatt*10+(*pc-48); pc++; }
      if ( iatt<0 ) fprintf( stderr,  "**** Error: negative value of xi=%d\n", iatt );
      pc++;
      term = pc;
      pc = strchr( pc, ' ' );
      fval = atof( term );
      if (iatt<=max_findex+1) W[iatt] = fval;
      //      printf( "W[%d] = %g\n", iatt, fval );

      if ( pc == NULL ) { break; }
      if ( pc-buf > bufsize ) { break; } /* end of line */
	//      if ( *(pc+1) == '\n') { break; }
      if ( *pc == ' ') { pc++; }
   } /* end of while */
   delete[] buf;

  ifs.close();

  return 1;
}

/*****************************************************************************/
/* read_hierarchy_file                                                       */
/*    (Arguments) fname: file name                                           */
/*    (Return)    #lines                                                     */
/*****************************************************************************/ 
int read_hierarchy_file ( char fname[] ) {

  /* local variables */
  std::ifstream ifs;  /* file descriptor */
  int i,j;            /* loop counter */
  char c;             /* character */
  char scat[100];     /* class string */
  std::string pcat;   /* parent*/
  std::string ccat;   /* child */
  int pcat_id;        /* parent*/
  int ccat_id;        /* child */
  int line = 0;       /* lines read */
  std::string strbuf; /* string buffer */
  char* buf;          /* char buffer */
  int bufsize;        /* char buffer size */

  /* Initialize */
  ifs.open( fname );
  if ( ifs.fail() ) { printf("File (%s) not found\n", fname); exit(1); }

  /* Feature value pairs */
  while ( std::getline( ifs,strbuf ) ) {
    bufsize = strbuf.size();
    buf = new char[bufsize+1];
    strcpy( buf, strbuf.c_str() );

    i=0;
    j=0;
    c = buf[i++];
    while ( c != ' ' ) {
          scat[j++] = c;
          c = buf[i++];
    }
    scat[j]='\0';
    //    pcat = atoi( scat );
    pcat = scat;
    if ( cat2i.find(pcat) != cat2i.end() ) {
      pcat_id = cat2i[pcat];
    } else {
      cat2i[pcat] = cat_serial;
      i2cat[cat_serial] = pcat;
      cat_serial++;
      pcat_id = cat2i[pcat];
    } 

    //    printf( "pcat = %d\n", pcat );

    j=0;
    c = buf[i++];
    while ( i <= bufsize ) {
      //    while ( c != '\n' ) {
          scat[j++] = c;
          c = buf[i++];
    }
    scat[j]='\0';
    //    ccat = atoi( scat );
    ccat = scat;
    if ( cat2i.find(ccat) != cat2i.end() ) {
      ccat_id = cat2i[ccat];
    } else {
      cat2i[ccat] = cat_serial;
      i2cat[cat_serial] = ccat;
      cat_serial++;
      ccat_id = cat2i[ccat];
    } 

    //    printf( "ccat = %d\n", ccat );

    child.insert(std::pair<int,int>(pcat_id,ccat_id));
    //    printf( "%d>%d\n", pcat, ccat );
    line++;
    delete[] buf;
  } /* end of while */
  //  children[count++] = 0;

  ifs.close();
  //  fclose( hfd );

  return line-1;
}

/*****************************************************************************/
/* read_test_file                                                       */
/*    (Arguments) fname: file name                                           */
/*    (Return)    #lines                                                     */
/*****************************************************************************/ 
int read_test_file ( char fname[] ) {

  /* Local variables */
  std::ifstream ifs;      /* file descriptor */
  int i,j;                /* loop counter */
  char c;                 /* child node */
  char satt[100];         /* attribute string */
  char sval[100];         /* value */
  char scat[100];         /* class */
  int icat;               /* class */
  int iatt;               /* class */
  float fval;             /* value */
  int datid = 0;          /* class */
  std::vector< std::pair< int, float > >* x;  /* data x */
  std::string strbuf;     /* string buffer */
  char* buf;              /* char buffer */
  int bufsize;            /* char buffer size */

  /* Initialize */
  ifs.open( fname );
  if ( ifs.fail() ) { printf("File (%s) not found\n", fname); exit(1); }

  /* Read lines */
  while ( std::getline( ifs,strbuf ) ) {
    bufsize = strbuf.size();
    buf = new char[bufsize+1];
    strcpy( buf, strbuf.c_str() );

    x = new std::vector< std::pair< int, float > >; 
    x->push_back( std::pair<int,float>(0, 1.0) );

    if ( buf[0]=='\0' ) { continue; }
    if ( buf[0]==EOF ) { return 0;}

    i=0; j=0; c = buf[i++];
    while ( c != ' ' ) { scat[j++] = c;  c = buf[i++];  }
    scat[j]='\0';
    icat = atoi( scat );
    icat++; /* dummy to avoid a waring */

    for( ; i<=bufsize ; ) {
        j=0;
        c = buf[i++];
       	while ( c != ':' ) { satt[j++] = c; c = buf[i++]; }
	satt[j]='\0';
        iatt = atoi( satt );
        if ( iatt<0 ) fprintf( stderr,  "**** Error: negatie value of xi=%d\n", iatt );
        if ( iatt>max_findex ) max_findex = iatt;

        j=0;
        c = buf[i++];
	while ( (c != ' ') && (c != '\n') && (c != '\0') && ( i<=bufsize ) ){
          sval[j++] = c;
          c = buf[i++];
        }
	sval[j]='\0';
        fval = atof( sval );
        if ( fval<0 ) fprintf( stderr,  "**** Error: negatie value of xv=%g\n", fval );
	fval = fval/(fval+1);
	// fval = 1/(1+exp(-fval));
        x->push_back( std::pair<int,float>(iatt,fval) );
	//         fprintf( stderr,  "(%d, %g)\n", iatt, fval );
    } /* end of for */
    //    printf( "\n" );
    datid++;
    if ( datid % 10000 == 0 ) {  fprintf( stderr, "*** %d data lines read\n", datid); }
    X.push_back( *x );

    delete[] buf;
  } /* end of while */

  ifs.close();
  return X.size();
}

/*****************************************************************************/
/* dot_prodct                                                                */
/*    (Arguments) command line                                               */
/*****************************************************************************/
float dot_product( int datid ) {
  float r=0;

  std::vector< std::pair< int, float > >::iterator iter;
  for( iter = X[datid].begin() ; iter != X[datid].end() ; iter++ ) {
    r += W[iter->first]*iter->second;
  }
  return r; 
  
}

/*****************************************************************************/
/* write_class                                                               */
/*    (Arguments) command line                                               */
/*****************************************************************************/
void write_class ( std::vector< int > *dat_list, char* fname ) {
  FILE *rfd;
  std::map< int, std::map< int,float > >::iterator iter;
  std::map< int,float >::iterator elem;
  std::map< int,float > catset;
  int i;
  int dcount;

  rfd = fopen( fname, "w" );
  if ( rfd==NULL ) { fprintf( stderr, "Cannot open file (%s)\n", fname); exit(1); }

  dcount = (*dat_list).size();
  for( i=0 ; i<dcount; i++ ) {
    int flg = 0;
    int d = (*dat_list)[i];
    iter = d2cat.find(d);
    catset = (*iter).second;
    for( elem = catset.begin(); elem != catset.end() ; elem++ ){
      flg = 1;
      if (((*elem).second)!=0.0) fprintf( rfd, "%s:%g ",i2cat[(*elem).first].c_str(), (*elem).second );
    }
    if ( flg ) fprintf( rfd, "\n" ); else fprintf( rfd, "%s:1\n", i2cat[root_id].c_str() );
  }
  fclose( rfd );
}

/*****************************************************************************/
/* usage                                                                     */
/*    (Arguments) command line                                               */
/*****************************************************************************/
void usage(char* command ) {
  fprintf( stderr, "   Usage: %s [options] 1> log1 2> log2\n",command );
  fprintf( stderr, "  options\n" );
  fprintf( stderr, "     -h <hierarchy file> (mandatory)\n" );
  fprintf( stderr, "     -t <test file> (mandatory)\n" );
  fprintf( stderr, "     -r <root ID> (default 0)\n" );
  fprintf( stderr, "     -m <model directory> (default \"EzeModel\")\n" );
  fprintf( stderr, "     -o <output file> (default result.eze)\n" );
}

/*****************************************************************************/
/* range01                                                                   */
/*    (Arguments) x: real value to convert                                   */
/*    (Return)    valune in the [0,1] range                                  */
/*****************************************************************************/
float range01 ( float x ) {
  if ( logistic ) return sigmoid(x);
  else return clip(x);
}

/*****************************************************************************/
/* clip                                                                      */
/*    (Arguments) x: real value to clip                                      */
/*    (Return)    valune in the [0,1] range                                  */
/*****************************************************************************/

float clip ( float x ) {
  if ( x > 1.0 ) return 1.0;
  if ( x < -1.0 ) return 0.0;
  return (x+1.0)/2.0;
}

/*****************************************************************************/
/* sigmoid                                                                   */
/*    (Arguments) x: real value to clip                                      */
/*    (Return)    valune in the [0,1] range                                  */
/*****************************************************************************/
 
float sigmoid ( float x ) {
  return 1/(1+exp(-x));
}

/****************************************************************************/
/*  read_option                                                             */
/*    (Arguments)  argc                                                     */
/*                 argv                                                     */
/****************************************************************************/ 

void read_option (int argc, char *argv[]) {
  char option_char;

  /* options */
  /*   -m <model directory> */
  /*   -h <hierarchy file> */
  /*   -r <root ID> */
  /*   -t <test file> */
  /*   -o <output file> */
  /*   -v (verbose) */

  while ((option_char = getopt(argc,argv,"m:h:r:d:t:o:v") ) != EOF) {
    switch (option_char) {  
      case 'm': strcpy(model_dir,optarg); break;
      case 'h': hierarchy_file = optarg; break;
      case 'r': root = optarg; break;
      case 't': test_file = optarg; break;
      case 'o': strcpy(result_file,optarg); break;
      case 'v': verbose = 1; break;
      case ':': usage( argv[0] ); exit(0);
      case '?': usage( argv[0] ); exit(0);
    }
  }
}
