/****************************************************************************/
/*                                                                          */ 
/* eze_induce (ver. 0.01): Efficient Hierarchical Classification Learner    */ 
/*                                                                          */ 
/* Author: Yutaka Sasaki, Toyota Technological Institute                    */ 
/*         (Yutaka.Sasaki<AT>toyota-ti.ac.jp)                               */ 
/*                                                                          */
/*   Release Date: March 27, 2016                                           */
/*                                                                          */
/*   Acknowledgment: This software development has been partly supported    */
/*                   by Japan's MEXT/JSPS KAKENHI Grant Number 25330271     */
/*                                                                          */ 
/* Licensed under the GNU General Public License, Version 2 (the "License") */
/* You may not use this file except in compliance with the License.         */
/* You may obtain a copy of the License at                                  */
/*                                                                          */
/*      http://http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt         */
/*                                                                          */
/* Unless required by applicable law or agreed to in writing, software      */
/* distributed under the License is distributed on an "AS IS" BASIS,        */
/* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. */
/* See the License for the specific language governing permissions and      */
/* limitations under the License.                                           */
/*                                                                          */ 
/****************************************************************************/

/* NB: This is not a good C++ code. This is C implementation with C++ data  */
/* type extension. Minimum Object-Oriented programming has been involved.   */

/* Header files */
#include <fstream>
#include <sstream>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctime>
#include <time.h>
#include <sys/time.h>
#include <map>
#include <set>
#include <vector>
#include <iterator>
#include "eze-0.01.h"

/* Global variables */
char version[] = "eze_induce (ver. 0.01): Hierarchical Classification System, Yutaka Sasaki, TTI, Japan.\n";
float* W;        /* SVM weight vector */
float* alpha;    /* SVM Lagrange multiplier */
float* Q;        /* SVM Qii = x_i \dot x_i*/

std::vector< std::vector< std::pair< int, float > > >  X;   /* example list */
std::vector< int >  positive;                               /* positives */
std::vector< int >  negative;                               /* negatives */
int num_dat = 0;      /* #all data */
int verbose=0;        /* command line option: verbose flag */
float C = DEFAULT_C;  /* command line option: SVM parameter */
int balanced = 0;     /* sampling strategy, 0: stochastic, 1: balanced */
int x_iteration = X_ITERATION;  /* #iterations */
char* data_file;      /* data file name */
const char* model_dir = MODEL; /* default model directory */
char* hierarchy_file; /* hierarchy file name */
int  max_findex = MAX_FINDEX; /* largest feature index */
int block = BLOCK_SIZE; /* default mini-batch size */
std::string root = ROOT; /* default root string */
int root_id = 0;         /* default root id */
float epsilon = EPSILON; /* default DCASVM stopping condition */

std::multimap< int, int > child;                /* class hierarchy child[p,c] */
std::map< int, std::map< int, bool > > closure; /* subclasses closure[p,c]=true/false */
std::map< int, std::map< int, bool > > c2dat;   /* class to data: c2dat[c,dat]=true/false */
std::map< int, std::map< int, bool > > d2cat;   /* dat to classes: d2cat[dat,c]=true/false */
std::map< int, std::string > i2cat;             /* category id to category string */
std::map< std::string, int > cat2i;             /* category string to category id */
int cat_serial = 1;                             /* category id */
bool* ccheck;                                   /* already explored or not */

/* Function prototypes */
void read_option( int argc, char *argv[] );
int read_hierarchy_file ( char fname[] );
int read_training_file ( char fname[] );
void propagate_ent( int root_id );
void propagate( int p, int node );
int train( int depth, int node, std::vector< int > *dat_list, std::vector< float > *val_list );
void usage( char* command );
float dot_product( int datid );
void print_class( std::vector< int > *v );
void write_W( char *fname );
void train_hierarchy( int level, int root_id, std::set< int > pool, std::map< int,float > score );
float range01 ( float x );
float clip ( float x );
float clipping ( float x, float L, float U );
double gettimeofday_sec();
float self_dot_product( int datid );
void dcal1_train( int max_iteration );

/* Main Program */
int main(int argc, char *argv[]) {

  /* Local variables */
  char* command;         /* invoked command */

  std::vector< int > dat_list;     /* initial dat list */
  std::vector< float > val_list;   /* initial value list */
  int result;                      /* file read status */
  std::set< int > root_samples;    /* samples at the root */
  std::map< int,float > score;     /* CCS scores */

  std::multimap< int, int >::iterator citer; /* iterator for classes */
  std::map< int, bool >::iterator diter;     /* iterator for data */
  std::map< int, bool >::iterator liter;     /* iterator for leaf classes */

  /* Read command line options */
  command = argv[0];
  read_option(argc,argv);
 
  if ( data_file == '\0' ) { usage(command); exit(1); }  
  if ( hierarchy_file == '\0' ) { usage(command); exit(1); }  

  /* Print header */
  printf( version );
  printf( "*** Hierarchy file name = %s\n", hierarchy_file );
  printf( "*** Training file name = %s\n", data_file );
  printf( "*** Model directory = %s\n", model_dir );
  printf( "*** Root ID = %s\n", root.c_str() );
  printf( "*** C = %g\n", C );
  printf( "*** DCASVM stopping criteria epsilon = %g\n", epsilon );
  printf( "*** DCASVM mini-batch size = %d\n", block );
  printf( "*** Sampling strategy (0:stochastic, 1:balanced) = %d\n", balanced );
  printf( "*** Verbose (0:silent, 1:verbose) = %d\n", verbose );
 
  /* Initialize */
  i2cat[0] = root;
  cat2i[root] = 0;
  srand((unsigned int)time(NULL));

  /* Read files */
  result = read_hierarchy_file ( hierarchy_file );
  fprintf( stderr, "#is_a in the hierarchy = %d\n",result );
  ccheck = new bool[cat_serial];

  num_dat = read_training_file ( data_file );
  fprintf( stderr, "%d data read.\n",num_dat );

  /* Create SVM parameter arrays */
  alpha = new float[num_dat];
  Q = new float[num_dat];
  alpha = new float[num_dat];
  for ( int k=0 ; k<num_dat ; k++ )  Q[k] = self_dot_product(k);

  /* Bottom-up category propagation  */
  for( int i=0; i< cat_serial ; i++ ) { ccheck[i] = false; }
  propagate_ent( root_id );
  fprintf( stderr, "Propagation done.\n" );

  /* Generate the example set for the root */
  for ( int i=0 ; i<num_dat ; i++ ) {
      root_samples.insert( i );
      score[i]=1.0;
  }

  /* Hierarchical training */
  W = new float[max_findex+2];
  for( int i=0; i< cat_serial ; i++ ) { ccheck[i] = false; }
  train_hierarchy( 1, root_id, root_samples, score );
  printf( "Training done.\n" );

  /* Clean up */
  delete[] W;
  delete[] alpha;
  delete[] Q;

} /* end of main */


/*****************************************************************************/
/*  train_hierarchy                                                          */
/*    (Arguments) level: current depth                                       */
/*                 node: target node                                         */
/*                 pool: a set of data                                       */
/*                 score: evaluation scores of data                          */
/*****************************************************************************/ 
void train_hierarchy ( int level, int node, std::set< int > pool, std::map< int,float > score ) {

  /* Local variables */
  int ccount;                /* # of child nodes */
  char model_file[NAME_LEN]; /* model file name */
  int c,d;                   /* loop variables */
  int pcount;                /* sample pool counter */

  std::map< int, float >* max_g = new std::map< int, float >;  /* max SVM reply */
  std::map< int, int >* max_child = new std::map< int, int >; /* and its class id */
  std::map< int, std::map< int, bool > >* cpositive = new std::map< int, std::map< int, bool > >;
                                                       /* positive data at a child node */

  std::map< int, std::set< int > > child_pool;         /* sample pool for a child node */
  std::map< int, std::map< int,float > > child_score;  /* ACCS scores for a child node */

  std::multimap< int, int >::iterator citer;    /* child iterator */
  std::map< int, bool >::iterator diter;        /* data iterator */
  std::map< int, bool >::iterator liter;        /* leaf class iterator */
  std::set< int >::iterator piter;              /* pool iterator */

  /* Check if already explored */
  if ( ccheck[node] ) { return; } /* return if already explored */ 
  ccheck[node] = true;

  fprintf( stderr, "Processing class %d.\n", node );
  ccount = child.count( node );
  fprintf( stderr, "(%d) %d child nodes found.\n",node, ccount );

  /* assign data to the leaf node */
  if ( ccount == 0) { 
        for( piter=pool.begin() ; piter!=pool.end() ; piter++ ) {
           d = (*piter);
           d2cat[d][node] = true;
        }
        return; 
  }

  /* max recorder initialization */
  for( piter=pool.begin() ; piter!=pool.end() ; piter++ ) {
        d = (*piter);
        (*max_g)[d] = -100.0;
  }

  /* set positives for each child node */
  (*cpositive).clear();
  citer=child.find( node ); 
  for ( int i=0 ; i<ccount ; i++ ) {
     c = citer->second;
     for ( liter=closure[c].begin() ; liter!=closure[c].end() ; liter++ ) {
       for ( diter=c2dat[liter->first].begin() ; diter!=c2dat[liter->first].end() ; diter++ ) {
	 (*cpositive)[c][diter->first] = true;
        }
     }
     citer++;
  } 

  /* train a model for each child node c */
  citer=child.find( node ); 
  child_pool.clear();
  for ( int i=0 ; i<ccount ; i++ ) {
    c = citer->second;
    sprintf( model_file, "%s/eze-%s-%s.model", model_dir, i2cat[node].c_str(), i2cat[c].c_str() );
    fprintf( stderr, "Model file name = %s\n",model_file );
    fprintf( stderr, "(%d/%d) Expanding %s-->%s (%d)\n", i, ccount, i2cat[node].c_str(), i2cat[c].c_str(), (int)pool.size() );

    std::map< int, bool > cc = (*cpositive)[c];

    positive.clear();
    negative.clear();
    for( int j=0 ; j<= max_findex+1 ; j++ ) { W[j] = 0.0; }
        
    pcount =0;
    for( piter=pool.begin() ; piter!=pool.end() ; piter++ ) {
       d = (*piter);
       if ( cc.find(d) != cc.end() ) { positive.push_back( d ); }
       else { negative.push_back( d ); }
       pcount++;
    }

    fprintf( stderr, "[%s] %d positives, %d negatives.\n", i2cat[c].c_str(), (int)positive.size(), (int)negative.size() );
    if ( positive.size()==0 ) { fprintf( stderr, "No positives found. Skipped.\n"); continue; }

    /* call DCASVM */
    dcal1_train( (int)( pool.size())*x_iteration  );

    /* generate sample pool for a child node */
    child_pool[c].clear(); 
    child_score[c].clear(); 
    for ( liter=closure[c].begin() ; liter!=closure[c].end() ; liter++ ) {
        for ( diter=c2dat[liter->first].begin() ; diter!=c2dat[liter->first].end() ; diter++ ) {
	    d = diter->first;
     	    child_pool[c].insert( d );
            child_score[c][d]= score[d]*range01(dot_product(d));
        }
    }

    /* write the weight vector of a child class */
    fprintf( stderr, "Writing model file \"%s\"\n", model_file );
    write_W( model_file );
    fprintf( stderr, "done.\n" );
    citer++;
  } /* end of for (child nodes) */

  /* clean up for memory efficiency */
  pool.clear();
  score.clear();
  delete cpositive;
  delete max_g;
  delete max_child;

  /* training at the child level */
  citer=child.find( node ); 
  for ( int i=0 ; i<ccount ; i++ ) {
     c = citer->second;
     /* move on to the training of a child node */
     train_hierarchy( level+1, c, child_pool[c], child_score[c] );
     fprintf( stderr, "(%d/%d) end\n", i+1, ccount );
     citer++;
  }
  //  ccheck[node] = false; /* revisit nodes if uncommented */
}

/*****************************************************************************/
/* dcal1_train                                                               */
/*    (Arguments) max_iteration: #max loop                                   */
/*****************************************************************************/ 
void dcal1_train( int max_iteration ) {
  int rnd;            /* random number */
  int d;              /* picked data */
  int y;              /* class label: +1, -1 */
  float UB = -1000.0; /* hard coding */
  float LB =  1000.0; /* hard coding */
  float G;            /* loss */ 
  float PG;
  float a;            /* old alpha[i] */
  int count=0;        /* loop counter */

  /* Header */
  fprintf( stderr, "<< Training in\n" );
  fprintf( stderr, "%d positives and %d negatives\n", (int)positive.size(), (int)negative.size() );
  if ( positive.size()+negative.size() == 0 ) return;

  /* Initialize */
  double start_time = gettimeofday_sec();
  for ( int k=0 ; k<num_dat ; k++ ) {  alpha[k] = 0.0; }

  /* Outer loop */
  while( count < max_iteration ) {
     UB = -1000.0; LB = 1000.0; /* hard coding */
     count++;
     /* Inner loop */
     for ( int t=0 ; t<block ; t++ ) {

        /* Sampling strategy */
        if (balanced) {
           if ( t % 2 == 0 ) { 
 	      if ( positive.size() == 0 ) continue;
              rnd = rand() % positive.size(); // random number in 0..#pdata-1
              d = positive[rnd];  
              y=1;
           } else {
 	      if ( negative.size() == 0 ) continue;
              rnd = rand() % negative.size();
              d = negative[rnd];
              y=-1;
           }
        } else {
           rnd = rand() % (positive.size()+negative.size());
           if ( rnd < (int)positive.size() ) {
              d = positive[rnd];  
              y=1;
           } else {
              d = negative[rnd-positive.size()];
              y=-1;
           }
        }

        /* loss and bound */
       G =  y*dot_product(d) - 1.0;
       PG = 0.0;
       if ( alpha[d] == 0.0 ) { 
          if ( G<0.0 ) PG = G; 
       }
       else if ( alpha[d] == C ) { 
          if ( G>0.0 ) PG = G; 
       } 
       else PG = G;

       if ( PG>UB ) UB=PG;
       if ( PG<LB ) LB=PG;

       /* update with L1 regularization */
       if ( PG != 0.0 ) {
         a = alpha[d];
         alpha[d] = clipping( alpha[d] - G/Q[d], 0.0, C );

         /* revise W */
         std::vector< std::pair< int, float > >::iterator xiter;
         for( xiter = X[d].begin() ; xiter != X[d].end() ; ++xiter ) {
            W[xiter->first] += (alpha[d] - a)*y*(xiter->second);
         } /* end of for */

       } /* end of if */  
     } /* end of for */

     if (verbose) fprintf( stderr, "UB %g,  LB %g, Diff %g, epsilon %g\n", UB, LB, UB-LB, epsilon );
     if ( (UB-LB < epsilon)&&(count>=MIN_ITERATION) ) {
       fprintf( stderr, ">> Training out (%g sec) (%d iterations)\n", gettimeofday_sec()-start_time, count );  
        return;
     }
  } /* end of while */
  fprintf( stderr, "Training out (%g sec) (%d iterations)\n", gettimeofday_sec()-start_time, count );  
}

/* clipping */
float clipping ( float x, float L, float U ) {
  if (x<L) return L;
  if (x>U) return U;
  return x;
}

/* get time */
double gettimeofday_sec() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + (double)tv.tv_usec*1e-6;
}

/* dot product between itself */
float self_dot_product( int datid ) {
  register float r=0;

  if ( (int)X.size() <= datid ) return 0.0;

  std::vector< std::pair< int, float > >::iterator xiter;
  for( xiter = X[datid].begin() ; xiter != X[datid].end() ; ++xiter ) {
    r += (xiter->second)*(xiter->second);
  }
  return r; 
}

/*****************************************************************************/
/* read_hierarchy_file                                                       */
/*    (Arguments) fname: file name                                           */
/*    (Return)    lines read                                                 */
/*****************************************************************************/ 
int read_hierarchy_file ( char fname[] ) {

  /* Local variables */
  std::ifstream ifs;                      /* file descriptor */
  int i,j;                                /* loop variables */
  char c;                                 /* character */
  char scat[NAME_LEN];                         /* strings of attribute, value, category */
  std::string pcat, ccat;                 /* parent/child categories, attribute, category */
  int pcat_id, ccat_id;                   /* parent/child category ids */
  int line = 0;                           /* counters */
  std::string strbuf;                     /* string buffer */
  char* buf;                              /* char buffer */
  int bufsize;                            /* char buffer size */

  ifs.open( fname );
  if ( ifs.fail() ) { printf("File (%s) not found\n", fname); exit(1); }

  /* file read loop */
  while ( std::getline( ifs,strbuf ) ) {
    bufsize = strbuf.size();
    buf = new char[bufsize+1];
    strcpy( buf, strbuf.c_str() );

    i=0; j=0;
    line++;
    c = buf[i++];
    while ( c != ' ' ) {
          scat[j++] = c;
          c = buf[i++];
    }
    scat[j]='\0'; 
    pcat = scat;
    if ( cat2i.find(pcat) != cat2i.end() ) {
      pcat_id = cat2i[pcat];
    } else {
      cat2i[pcat] = cat_serial;
      i2cat[cat_serial] = pcat;
      cat_serial++;
      pcat_id = cat2i[pcat];
    } 

    j=0;
    c = buf[i++];
    while ( i <= bufsize ) {
          scat[j++] = c;
          c = buf[i++];
    }
    scat[j]='\0';
    ccat = scat;
    if ( cat2i.find(ccat) != cat2i.end() ) {
      ccat_id = cat2i[ccat];
    } else {
      cat2i[ccat] = cat_serial;
      i2cat[cat_serial] = ccat;
      cat_serial++;
      ccat_id = cat2i[ccat];
    } 
    //    printf( "%d > %d\n", pcat, ccat );  // for debugging

    child.insert(std::pair<int,int>(pcat_id,ccat_id));

    delete[] buf;
  } /* end of while */

  ifs.close();

  return line-1;
}

/*****************************************************************************/
/* read_training_file                                                        */
/*    (Arguments) fname: file name                                           */
/*    (Return)    # data                                                     */
/*****************************************************************************/ 
int read_training_file ( char fname[] ) {

  /* Local variables */
  std::ifstream ifs;         /* file descriptor */
  int iatt, icat_id=1;       /* attribute, category */
  std::string icat;          /* attribute, category */
  register char *term;       /* term */
  float fval;                /* feature value */
  int datid = 0;             /* dat id */
  int fcount = 0;            /* feature counter */
  register char *pc;         /* pointer */
  char *stat;                /* pointer */
  std::vector< std::pair< int, float > >* x;  /* data x */
  std::string strbuf;        /* string buffer */
  char* buf;                 /* char buffer */
  int bufsize;               /* char buffer size */

  /* File read loop */
  ifs.open( fname );
  if ( ifs.fail() ) { printf("File (%s) not found\n", fname); exit(1); }

  while ( std::getline( ifs,strbuf ) ) {
    //    fprintf( stderr,  "%s\n", strbuf.c_str() );
    bufsize = strbuf.size();
    buf = new char[bufsize+1];

    strcpy( buf, strbuf.c_str() );
    if ( buf[0] == '\0' ) { datid++; continue; }

    x = new std::vector< std::pair< int, float > >; 

    pc = buf;
    x->push_back( std::pair<int,float>(0, 1.0) );

    while ( 1 ) {
       term = pc;
       stat = strchr( pc, ',' );
       if ( stat == NULL ) { 
           pc = strchr( pc, ' ' );
           *pc = '\0';
           pc++;
           icat = term;

          if ( cat2i.find(icat) != cat2i.end() ) {
             icat_id = cat2i[icat];
          } else {
             cat2i[icat] = cat_serial;
             i2cat[cat_serial] = icat;
             cat_serial++;
             icat_id = cat2i[icat];
          } 

           c2dat[icat_id][datid] = true;
	   if ( !closure[icat_id][icat_id] ) closure[icat_id][icat_id] = true;
           break; 

       } else {

           pc = stat;
           *pc = '\0';
           icat = term;
	   pc++;
           while ( *pc==' ' ) pc++;

           if ( cat2i.find(icat) != cat2i.end() ) {
             icat_id = cat2i[icat];
           } else {
             cat2i[icat] = cat_serial;
             i2cat[cat_serial] = icat;
             cat_serial++;
             icat_id = cat2i[icat];
           } 

           c2dat[icat_id][datid] = true;
           if ( !closure[icat_id][icat_id] ) closure[icat_id][icat_id] = true;

       }
    }

    /* Read feature list */
    fcount=0; 
    while ( *pc != '\0' ) {
        pc++;
        iatt=0;
        while ( *pc != ':' ) { iatt = iatt*10+(*pc-48); pc++; }
        if ( iatt<0 ) fprintf( stderr,  "**** Error: negative value of xi=%d\n", iatt );
        if ( iatt > max_findex ) max_findex = iatt;
        pc++;
        term = pc;
        pc = strchr( pc, ' ' );
        fcount++;
        if ( pc == NULL ) {   
            fval = atof( term );
            // fval = 1/(1+exp(-fval));
	    fval = fval/(fval+1);
            x->push_back( std::pair<int,float>(iatt,fval) );
	    //	    fprintf( stderr,  "NULL(%d, %g)\n", iatt, fval );
            break;
        } else {
            fval = atof( term );
            // fval = 1/(1+exp(-fval));
	    fval = fval/(fval+1);
            x->push_back( std::pair<int,float>(iatt,fval) );
	    //	    fprintf( stderr,  "(%d, %g)\n", iatt, fval );
	}
        if ( pc-buf > bufsize ) { break; }
    } /* end of while */
    //    fprintf( stderr, "*** (%d) %d features\n", datid, fcount );

    datid++;
    if ( datid % 10000 == 0 ) {  fprintf( stderr, "*** %d data lines read\n", datid); }
     X.push_back( *x );

     delete[] buf;
  } /* end of while */

  fprintf( stderr, "cat_serial = %d\n", cat_serial );
  ifs.close();
  return X.size();
}

/*****************************************************************************/
/* usage                                                                     */
/*    (Arguments) command line                                               */
/*****************************************************************************/
void usage(char* command ) {
  fprintf( stderr, " %s", version );
  fprintf( stderr, "   Usage: %s [options] 1> log1 2> log2\n",command );
  fprintf( stderr, "    options\n" );
  fprintf( stderr, "     -h <hierarchy file> (mandatory)\n" );
  fprintf( stderr, "     -r <root ID> (default 0)\n" );
  fprintf( stderr, "     -d <data file> (mandatory)\n" );
  fprintf( stderr, "     -m <model directory> (default \"EzeModel\")\n" );
  fprintf( stderr, "     -e <epsilon> (for DCASVM, default 0.5)\n" );
  fprintf( stderr, "     -k <block size> (for DCASVM, default 100)\n" );
  fprintf( stderr, "     -v (verbose)\n" );
  fprintf( stderr, "     -b (balanced sampling)\n" );
  fprintf( stderr, "     -x <# of fold>\n" );

}

/*****************************************************************************/
/* dot_prodct                                                                */
/*    (Arguments) command line                                               */
/*****************************************************************************/
float dot_product( int datid ) {
  register float r=0;

  std::vector< std::pair< int, float > >::iterator xiter;
  for( xiter = X[datid].begin() ; xiter != X[datid].end() ; ++xiter ) {
    r += W[(*xiter).first]*((*xiter).second);
  }
  return r; 
}

/*****************************************************************************/
/* propagate_ent                                                             */
/*    (Arguments) root node id                                               */
/*****************************************************************************/
void propagate_ent( int top ) {
    /* local variables */
    std::multimap< int, int >::iterator citer;
    int ccount = child.count( top );
    fprintf( stderr, "*** %d children for root found\n", ccount );

     citer = child.find( top ); 
     for ( int i=0 ; i<ccount ; i++ ) {
         int c = citer->second;
         propagate( top, c );
         citer++; 
     }
     fprintf( stderr, "*** %d leaf nodes assigned to root\n", (int)closure[top].size() );
}

/*****************************************************************************/
/* propagate                                                                 */
/*    (Arguments) parent node id                                             */
/*                target  node id                                            */
/*****************************************************************************/
void propagate( int p, int node ) {
     /* local variables */
     std::multimap< int, int >::iterator citer; 
     std::multimap< int, std::pair< int,int > >::iterator diter;
     std::multimap< int, int >::iterator piter;
     std::map< int, bool >::iterator iter;

     if ( !ccheck[node] ) {  /* not explored yet */
        ccheck[node] = true;
        int ccount = child.count( node );
	//        fprintf( stderr, "*** %d children for %d found\n", ccount, node );

        citer = child.find( node ); 
        for ( int i=0 ; i<ccount ; i++ ) {
            int c = citer->second;
	    //            fprintf( stderr, "*** Expanding %d --> %d.\n",node,c );
            propagate( node, c );
            citer++; 
        }
     } 

     for ( iter=closure[node].begin() ; iter!=closure[node].end(); iter++ ) {
       closure[p][iter->first] = true;
      }
}

/*****************************************************************************/
/* write_W                                                                   */
/*    (Arguments) fname: file name                                           */
/*****************************************************************************/
void write_W ( char *fname ) {

  FILE *fp = fopen( fname, "w" );
  if ( fp==NULL ) { printf("Cannot open file (%s)\n", fname); exit(1); }

  fprintf( fp, "%g", W[0] );
   for( int j=1; j<= max_findex+1 ; j++ ) { 
      if ( W[j] != 0.0 ) { fprintf( fp, " %d:%g", j, W[j] ); }
   }
   fprintf( fp, "\n" );
   fclose( fp );
}

/*****************************************************************************/
/* range01                                                                   */
/*    (Arguments) x: real value to clip                                      */
/*    (Return)    value in the [0,1] range                                   */
/*****************************************************************************/
float range01 ( float x ) {
 return clip(x);
}

/*****************************************************************************/
/* clip                                                                      */
/*    (Arguments) x: real value to clip                                      */
/*    (Return)    value in the [0,1] range                                   */
/*****************************************************************************/

float clip ( float x ) {
  if ( x>1.0 ) return 1.0;
  if ( x<-1.0 ) return 0.0;
  return (x+1)/2;
}

/****************************************************************************/
/*  read_option                                                             */
/*    (Arguments)  argc                                                     */
/*                 argv                                                     */
/****************************************************************************/ 

void read_option (int argc, char *argv[]) {
  char option_char;

  /* options */
  /*   -m <model directory> */
  /*   -h <hierarchy file> */
  /*   -r <root ID> */
  /*   -d <data file> */
  /*   -e <epsilon> (DCASVM stopping condition) */
  /*   -k <block size> (DCASVM mini-batch size) */
  /*   -i <iterations> (magnification to #data) */
  /*   -c <SVM hyper-parameter> (default 1.0) */
  /*   -v (verbose) */
  /*   -b (balanced sampling) */
  /*   -h (show this help) */

  while ((option_char = getopt(argc,argv,"m:h:r:d:i:k:c:e:vb") ) != EOF) {
    switch (option_char) {  
      case 'm': model_dir = optarg; break;
      case 'h': hierarchy_file = optarg; break;
      case 'r': root = optarg; break;
      case 'd': data_file = optarg; break;
      case 'i': x_iteration  = atoi(optarg); break;
      case 'k': block = atoi(optarg); break;
      case 'c': C = atof(optarg); 
           if (C>0.0) break; 
           else { fprintf( stderr, "C should be more than zero.\n"); exit(0); }
      case 'e': epsilon = atof(optarg); break;
      case 'v': verbose = 1; break;
      case 'b': balanced = 1; break;
      case ':': usage( argv[0] ); exit(0);
      case '?': usage( argv[0] ); exit(0);
    }
  }
}

/*****************************************************************************/

